import logging
from datetime import datetime

import sqlalchemy as sa
from dateutil.tz import tzutc
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common import redis_helper
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import ResourceSlot, SessionResult, SessionTypes

from ..models import (
    DefaultForUnspecified,
    DomainRow,
    GroupRow,
    KeyPairResourcePolicyRow,
    KeyPairRow,
    SessionDependencyRow,
    SessionRow,
    UserRow,
)
from ..models.utils import execute_with_retry
from .types import PredicateResult, SchedulingContext

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.scheduler"))

_check_keypair_concurrency_script = """
local key = KEYS[1]
local limit = tonumber(ARGV[1])
local result = {}
redis.call('SETNX', key, 0)
local count = tonumber(redis.call('GET', key))
if limit > 0 and count >= limit then
    result[1] = 0
    result[2] = count
    return result
end
redis.call('INCR', key)
result[1] = 1
result[2] = count + 1
return result
"""


async def check_reserved_batch_session(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    """
    Check if a batch-type session should not be started for a certain amount of time.
    """
    if sess_ctx.session_type == SessionTypes.BATCH:
        query = sa.select(SessionRow.starts_at).where(SessionRow.id == sess_ctx.id)
        starts_at = await db_sess.scalar(query)
        if starts_at is not None and datetime.now(tzutc()) < starts_at:
            return PredicateResult(
                False,
                "Before start time",
            )
    return PredicateResult(True)


async def check_concurrency(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    async def _get_max_concurrent_sessions() -> int:
        resouce_policy_q = sa.select(KeyPairRow.resource_policy).where(
            KeyPairRow.access_key == sess_ctx.access_key
        )
        if sess_ctx.is_private:
            concurrent_session_column = KeyPairResourcePolicyRow.max_concurrent_sftp_sessions
        else:
            concurrent_session_column = KeyPairResourcePolicyRow.max_concurrent_sessions
        select_query = sa.select(concurrent_session_column).where(
            KeyPairResourcePolicyRow.name == resouce_policy_q.scalar_subquery()
        )
        result = await db_sess.execute(select_query)
        return result.scalar()

    max_concurrent_sessions = await execute_with_retry(_get_max_concurrent_sessions)
    if sess_ctx.is_private:
        redis_key = f"keypair.sftp_concurrency_used.{sess_ctx.access_key}"
    else:
        redis_key = f"keypair.concurrency_used.{sess_ctx.access_key}"
    ok, concurrency_used = await redis_helper.execute_script(
        sched_ctx.registry.redis_stat,
        "check_keypair_concurrency_used",
        _check_keypair_concurrency_script,
        [redis_key],
        [max_concurrent_sessions],
    )
    if ok == 0:
        return PredicateResult(
            False,
            f"You cannot run more than {max_concurrent_sessions} concurrent sessions",
        )
    log.debug(
        "number of concurrent sessions of ak:{0} = {1} / {2}",
        sess_ctx.access_key,
        concurrency_used,
        max_concurrent_sessions,
    )
    return PredicateResult(True)


async def check_dependencies(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    j = sa.join(
        SessionDependencyRow,
        SessionRow,
        SessionDependencyRow.depends_on == SessionRow.id,
    )
    query = (
        sa.select(
            SessionRow.id,
            SessionRow.name,
            SessionRow.result,
        )
        .select_from(j)
        .where(SessionDependencyRow.session_id == sess_ctx.id)
    )
    result = await db_sess.execute(query)
    rows = result.fetchall()
    pending_dependencies = []
    for row in rows:
        if row.result != SessionResult.SUCCESS:
            pending_dependencies.append(row)
    all_success = not pending_dependencies
    if all_success:
        return PredicateResult(True)
    return PredicateResult(
        False,
        "Waiting dependency sessions to finish as success. ({})".format(
            ", ".join(f"{row.name} ({row.id})" for row in pending_dependencies),
        ),
    )


async def check_keypair_resource_limit(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    resouce_policy_q = sa.select(KeyPairRow.resource_policy).where(
        KeyPairRow.access_key == sess_ctx.access_key
    )
    select_query = sa.select(KeyPairResourcePolicyRow).where(
        KeyPairResourcePolicyRow.name == resouce_policy_q.scalar_subquery()
    )
    result = await db_sess.execute(select_query)
    resource_policy = result.scalars().first()
    resource_policy_map = {
        "total_resource_slots": resource_policy.total_resource_slots,
        "default_for_unspecified": resource_policy.default_for_unspecified,
    }
    total_keypair_allowed = ResourceSlot.from_policy(
        resource_policy_map, sched_ctx.known_slot_types
    )
    key_occupied = await sched_ctx.registry.get_keypair_occupancy(
        sess_ctx.access_key, db_sess=db_sess
    )
    log.debug("keypair:{} current-occupancy: {}", sess_ctx.access_key, key_occupied)
    log.debug("keypair:{} total-allowed: {}", sess_ctx.access_key, total_keypair_allowed)
    if not (key_occupied + sess_ctx.requested_slots <= total_keypair_allowed):
        return PredicateResult(
            False,
            "Your keypair resource quota is exceeded. ({})".format(
                " ".join(
                    f"{k}={v}"
                    for k, v in total_keypair_allowed.to_humanized(
                        sched_ctx.known_slot_types
                    ).items()
                )
            ),
        )
    return PredicateResult(True)


async def check_user_resource_limit(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    main_ak = (
        sa.select(UserRow.main_access_key)
        .where(UserRow.uuid == sess_ctx.user_uuid)
        .scalar_subquery()
    )
    resouce_policy_q = sa.select(KeyPairRow.resource_policy).where(KeyPairRow.access_key == main_ak)
    select_query = sa.select(KeyPairResourcePolicyRow).where(
        KeyPairResourcePolicyRow.name == resouce_policy_q.scalar_subquery()
    )
    resource_policy: KeyPairResourcePolicyRow | None = (await db_sess.scalars(select_query)).first()
    if resource_policy is None:
        return PredicateResult(
            False,
            f"User has no main-keypair or the main-keypair has no keypair resource policy (uid: {sess_ctx.user_uuid})",
        )

    resource_policy_map = {
        "total_resource_slots": resource_policy.total_resource_slots,
        "default_for_unspecified": resource_policy.default_for_unspecified,
    }
    total_main_keypair_allowed = ResourceSlot.from_policy(
        resource_policy_map, sched_ctx.known_slot_types
    )
    user_occupied = await sched_ctx.registry.get_user_occupancy(sess_ctx.user_uuid, db_sess=db_sess)
    log.debug("user:{} current-occupancy: {}", sess_ctx.user_uuid, user_occupied)
    log.debug("user:{} total-allowed: {}", sess_ctx.user_uuid, total_main_keypair_allowed)
    if not (user_occupied + sess_ctx.requested_slots <= total_main_keypair_allowed):
        return PredicateResult(
            False,
            "Your main-keypair resource quota is exceeded. ({})".format(
                " ".join(
                    f"{k}={v}"
                    for k, v in total_main_keypair_allowed.to_humanized(
                        sched_ctx.known_slot_types
                    ).items()
                )
            ),
        )
    return PredicateResult(True)


async def check_group_resource_limit(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    query = sa.select(GroupRow.total_resource_slots).where(GroupRow.id == sess_ctx.group_id)
    group_resource_slots = await db_sess.scalar(query)
    group_resource_policy = {
        "total_resource_slots": group_resource_slots,
        "default_for_unspecified": DefaultForUnspecified.UNLIMITED,
    }
    total_group_allowed = ResourceSlot.from_policy(
        group_resource_policy, sched_ctx.known_slot_types
    )
    group_occupied = await sched_ctx.registry.get_group_occupancy(
        sess_ctx.group_id, db_sess=db_sess
    )
    log.debug("group:{} current-occupancy: {}", sess_ctx.group_id, group_occupied)
    log.debug("group:{} total-allowed: {}", sess_ctx.group_id, total_group_allowed)
    if not (group_occupied + sess_ctx.requested_slots <= total_group_allowed):
        return PredicateResult(
            False,
            "Your group resource quota is exceeded. ({})".format(
                " ".join(
                    f"{k}={v}"
                    for k, v in total_group_allowed.to_humanized(sched_ctx.known_slot_types).items()
                )
            ),
        )
    return PredicateResult(True)


async def check_domain_resource_limit(
    db_sess: SASession,
    sched_ctx: SchedulingContext,
    sess_ctx: SessionRow,
) -> PredicateResult:
    query = sa.select(DomainRow.total_resource_slots).where(DomainRow.name == sess_ctx.domain_name)
    domain_resource_slots = await db_sess.scalar(query)
    domain_resource_policy = {
        "total_resource_slots": domain_resource_slots,
        "default_for_unspecified": DefaultForUnspecified.UNLIMITED,
    }
    total_domain_allowed = ResourceSlot.from_policy(
        domain_resource_policy, sched_ctx.known_slot_types
    )
    domain_occupied = await sched_ctx.registry.get_domain_occupancy(
        sess_ctx.domain_name, db_sess=db_sess
    )
    log.debug("domain:{} current-occupancy: {}", sess_ctx.domain_name, domain_occupied)
    log.debug("domain:{} total-allowed: {}", sess_ctx.domain_name, total_domain_allowed)
    if not (domain_occupied + sess_ctx.requested_slots <= total_domain_allowed):
        return PredicateResult(
            False,
            "Your domain resource quota is exceeded. ({})".format(
                " ".join(
                    f"{k}={v}"
                    for k, v in total_domain_allowed.to_humanized(
                        sched_ctx.known_slot_types
                    ).items()
                )
            ),
        )
    return PredicateResult(True)



from __future__ import annotations

import sys
from decimal import Decimal
from typing import List, Optional, Sequence, Tuple

import trafaret as t

from ai.backend.common.types import (
    AgentId,
    AgentSelectionStrategy,
    ResourceSlot,
    SessionId,
)

from ..models import AgentRow, SessionRow
from .types import AbstractScheduler, KernelInfo


def get_slot_index(slotname: str, agent_selection_resource_priority: list[str]) -> int:
    try:
        return agent_selection_resource_priority.index(slotname)
    except ValueError:
        return sys.maxsize


def key_by_remaining_slots(
    agent: AgentRow,
    requested_slots: ResourceSlot,
    agent_selection_strategy: AgentSelectionStrategy,
    agent_selection_resource_priority: list[str],
) -> Tuple[int, ...]:
    unused_slot_keys = set()
    for k, v in requested_slots.items():
        if v == Decimal(0):
            unused_slot_keys.add(k)
    num_extras = 0
    for k, v in agent.available_slots.items():
        if k in unused_slot_keys and v > Decimal(0):
            num_extras += 1

    for requested_slot_key in sorted(requested_slots.data.keys(), reverse=True):
        device_name = requested_slot_key.split(".")[0]
        if (
            requested_slot_key not in agent_selection_resource_priority
            and device_name in agent_selection_resource_priority
        ):
            agent_selection_resource_priority.insert(
                agent_selection_resource_priority.index(device_name) + 1, requested_slot_key
            )

    resource_priorities = sorted(
        requested_slots.data.keys(),
        key=lambda item: get_slot_index(item, agent_selection_resource_priority),
    )

    remaining_slots = agent.available_slots - agent.occupied_slots

    # If the requested slot does not exist in the corresponding agent,
    # the agent should not be selected, in this case it puts -math.inf for avoiding to being selected.
    match agent_selection_strategy:
        case AgentSelectionStrategy.LEGACY:
            comparators = [
                agent.available_slots.get(key, -sys.maxsize) for key in resource_priorities
            ]
        case AgentSelectionStrategy.CONCENTRATED:
            comparators = [-remaining_slots.get(key, sys.maxsize) for key in resource_priorities]
        case AgentSelectionStrategy.DISPERSED | _:
            comparators = [remaining_slots.get(key, -sys.maxsize) for key in resource_priorities]

    # Put back agents with more extra slot types
    # (e.g., accelerators)
    # Also put front agents with exactly required slot types
    return (-num_extras, *comparators)


class FIFOSlotScheduler(AbstractScheduler):
    config_iv = t.Dict({
        t.Key("num_retries_to_skip", default=0): t.ToInt(gte=0),
    }).allow_extra("*")

    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[SessionRow],
        existing_sessions: Sequence[SessionRow],
    ) -> Optional[SessionId]:
        local_pending_sessions = list(pending_sessions)
        skipped_sessions: List[SessionRow] = []
        max_retries = self.config["num_retries_to_skip"]
        while local_pending_sessions:
            # Just pick the first pending session, but skip it
            # if it has more than 3 failures.
            s = local_pending_sessions.pop(0)
            if max_retries == 0:  # it's strict FIFO
                return s.id
            if s.status_data is not None:
                sched_data = s.status_data.get("scheduler", {})
                if sched_data.get("retries", 0) >= max_retries:
                    skipped_sessions.append(s)
                    continue
            return s.id
        # But if all sessions are skipped, then choose the first one.
        if skipped_sessions:
            return skipped_sessions[0].id
        return None

    def _assign_agent(
        self,
        agents: Sequence[AgentRow],
        requested_slots: ResourceSlot,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        possible_agents = []
        for agent in agents:
            remaining_slots = agent.available_slots - agent.occupied_slots
            if remaining_slots >= requested_slots:
                possible_agents.append(agent)
        if possible_agents:
            chosen_agent = max(
                possible_agents,
                key=lambda agent: key_by_remaining_slots(
                    agent,
                    requested_slots,
                    agent_selection_strategy,
                    agent_selection_resource_priority,
                ),
            )
            return chosen_agent.id
        return None

    def assign_agent_for_session(
        self,
        agents: Sequence[AgentRow],
        pending_session: SessionRow,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_session.requested_slots,
            agent_selection_strategy,
            agent_selection_resource_priority,
        )

    def assign_agent_for_kernel(
        self,
        agents: Sequence[AgentRow],
        pending_kernel: KernelInfo,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_kernel.requested_slots,
            agent_selection_strategy,
            agent_selection_resource_priority,
        )


class LIFOSlotScheduler(AbstractScheduler):
    config_iv = t.Dict({}).allow_extra("*")

    def pick_session(
        self,
        total_capacity: ResourceSlot,
        pending_sessions: Sequence[SessionRow],
        existing_sessions: Sequence[SessionRow],
    ) -> Optional[SessionId]:
        # Just pick the last pending session.
        return SessionId(pending_sessions[-1].id)

    def _assign_agent(
        self,
        agents: Sequence[AgentRow],
        requested_slots: ResourceSlot,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        possible_agents = []
        for agent in agents:
            remaining_slots = agent.available_slots - agent.occupied_slots
            if remaining_slots >= requested_slots:
                possible_agents.append(agent)
        if possible_agents:
            chosen_agent = max(
                possible_agents,
                key=lambda agent: key_by_remaining_slots(
                    agent,
                    requested_slots,
                    agent_selection_strategy,
                    agent_selection_resource_priority,
                ),
            )
            return chosen_agent.id
        return None

    def assign_agent_for_session(
        self,
        agents: Sequence[AgentRow],
        pending_session: SessionRow,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_session.requested_slots,
            agent_selection_strategy,
            agent_selection_resource_priority,
        )

    def assign_agent_for_kernel(
        self,
        agents: Sequence[AgentRow],
        pending_kernel: KernelInfo,
        agent_selection_strategy: AgentSelectionStrategy,
        agent_selection_resource_priority: list[str],
    ) -> Optional[AgentId]:
        return self._assign_agent(
            agents,
            pending_kernel.requested_slots,
            agent_selection_strategy,
            agent_selection_resource_priority,
        )



from __future__ import annotations

import asyncio
import logging
import time
from contextlib import asynccontextmanager as actxmgr
from contextvars import ContextVar
from typing import AsyncIterator, Callable, Optional

import sqlalchemy as sa
import zmq
from callosum.lower.zeromq import ZeroMQAddress, ZeroMQRPCTransport
from callosum.rpc import Peer, RPCUserError
from sqlalchemy.engine.row import Row

from ai.backend.common import msgpack
from ai.backend.common.auth import ManagerAuthHandler, PublicKey, SecretKey
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import AgentId

from .api.exceptions import (
    AgentError,
)
from .models.agent import agents
from .models.utils import ExtendedAsyncSAEngine, execute_with_retry

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class PeerInvoker(Peer):
    class _CallStub:
        _cached_funcs: dict[str, Callable]
        order_key: ContextVar[Optional[str]]

        def __init__(self, peer: Peer):
            self._cached_funcs = {}
            self.peer = peer
            self.order_key = ContextVar("order_key", default=None)

        def __getattr__(self, name: str):
            if f := self._cached_funcs.get(name, None):
                return f
            else:

                async def _wrapped(*args, **kwargs):
                    request_body = {
                        "args": args,
                        "kwargs": kwargs,
                    }
                    self.peer.last_used = time.monotonic()
                    ret = await self.peer.invoke(name, request_body, order_key=self.order_key.get())
                    self.peer.last_used = time.monotonic()
                    return ret

                self._cached_funcs[name] = _wrapped
                return _wrapped

    call: _CallStub
    last_used: float

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.call = self._CallStub(self)
        self.last_used = time.monotonic()


class AgentRPCCache:
    _cache: dict[AgentId, tuple[str, PublicKey | None]]

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        manager_public_key: PublicKey,
        manager_secret_key: SecretKey,
        *,
        rpc_keepalive_timeout: int = 60,  # in seconds
    ) -> None:
        self.db = db
        self.manager_public_key = manager_public_key
        self.manager_secret_key = manager_secret_key
        self.rpc_keepalive_timeout = rpc_keepalive_timeout
        self._cache = {}

    def update(
        self,
        agent_id: AgentId,
        agent_addr: str,
        public_key: Optional[PublicKey],
    ) -> None:
        self._cache[agent_id] = (agent_addr, public_key)

    def discard(self, agent_id: AgentId) -> None:
        self._cache.pop(agent_id, None)

    async def get_rpc_args(self, agent_id: AgentId) -> tuple[str, PublicKey | None]:
        cached_args = self._cache.get(agent_id, None)
        if cached_args:
            return cached_args

        async def _fetch_agent() -> Row:
            async with self.db.begin_readonly() as conn:
                query = (
                    sa.select([agents.c.addr, agents.c.public_key])
                    .select_from(agents)
                    .where(
                        agents.c.id == agent_id,
                    )
                )
                result = await conn.execute(query)
                return result.first()

        agent = await execute_with_retry(_fetch_agent)
        return agent["addr"], agent["public_key"]

    @actxmgr
    async def rpc_context(
        self,
        agent_id: AgentId,
        *,
        invoke_timeout: Optional[float] = None,
        order_key: Optional[str] = None,
    ) -> AsyncIterator[PeerInvoker]:
        agent_addr, agent_public_key = await self.get_rpc_args(agent_id)
        keepalive_retry_count = 3
        keepalive_interval = self.rpc_keepalive_timeout // keepalive_retry_count
        if keepalive_interval < 2:
            keepalive_interval = 2
        if agent_public_key:
            auth_handler = ManagerAuthHandler(
                "local",
                agent_public_key,
                self.manager_public_key,
                self.manager_secret_key,
            )
        else:
            auth_handler = None
        log.debug(
            "rpc_context(): calling ag:{} via {}, with agent_public_key:{!r}",
            agent_id,
            agent_addr,
            agent_public_key.decode() if agent_public_key else None,
        )
        peer = PeerInvoker(
            connect=ZeroMQAddress(agent_addr),
            transport=ZeroMQRPCTransport,
            authenticator=auth_handler,
            transport_opts={
                "zsock_opts": {
                    zmq.TCP_KEEPALIVE: 1,
                    zmq.TCP_KEEPALIVE_IDLE: self.rpc_keepalive_timeout,
                    zmq.TCP_KEEPALIVE_INTVL: keepalive_interval,
                    zmq.TCP_KEEPALIVE_CNT: keepalive_retry_count,
                },
            },
            serializer=msgpack.packb,
            deserializer=msgpack.unpackb,
        )
        try:
            async with asyncio.timeout(invoke_timeout), peer:
                okey_token = peer.call.order_key.set("")
                try:
                    yield peer
                finally:
                    peer.call.order_key.reset(okey_token)
        except RPCUserError as orig_exc:
            raise AgentError(agent_id, orig_exc.name, orig_exc.repr, orig_exc.args)
        except Exception:
            raise



"""add_auto_terminate_col

Revision ID: 857bdec5abda
Revises: 210c4d9be768
Create Date: 2023-05-23 12:18:02.046409

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "857bdec5abda"
down_revision = "210c4d9be768"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "agents",
        sa.Column(
            "auto_terminate_abusing_kernel",
            sa.Boolean(),
            server_default=sa.text("false"),
            nullable=False,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("agents", "auto_terminate_abusing_kernel")
    # ### end Alembic commands ###



import logging
import re
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, Iterable, Tuple

import aiohttp
import aiohttp_cors
import aiotools
import attrs
import sqlalchemy as sa
import trafaret as t
import yaml
from aiohttp import web
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.exc import NoResultFound

from ai.backend.common import validators as tx
from ai.backend.common.config import model_definition_iv
from ai.backend.common.docker import ImageRef
from ai.backend.common.events import KernelLifecycleEventReason
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import ClusterMode, SessionTypes, VFolderID, VFolderUsageMode
from ai.backend.manager.registry import check_scaling_group

from ..defs import DEFAULT_CHUNK_SIZE, DEFAULT_IMAGE_ARCH
from ..models import (
    EndpointLifecycle,
    EndpointRow,
    EndpointTokenRow,
    ImageRow,
    RouteStatus,
    RoutingRow,
    UserRow,
    query_accessible_vfolders,
    resolve_group_name_or_id,
    scaling_groups,
    vfolders,
)
from ..types import UserScope
from .auth import auth_required
from .exceptions import InvalidAPIParameters, ObjectNotFound, ServiceUnavailable, VFolderNotFound
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from .session import query_userinfo
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params, get_access_key_scopes, get_user_uuid_scopes, undefined

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__name__))


class UndefChecker(t.Trafaret):
    def check_and_return(self, value: Any) -> object:
        if value == undefined:
            return value
        else:
            self._failure("Invalid Undef format", value=value)
            return None


async def is_user_allowed_to_access_resource(
    db_sess: AsyncSession,
    request: web.Request,
    resource_owner: uuid.UUID,
) -> bool:
    if request["user"]["is_superadmin"]:
        return True
    elif request["user"]["is_admin"]:
        query = sa.select(UserRow).filter(UserRow.uuid == resource_owner)
        result = await db_sess.execute(query)
        user = result.scalar()
        return user.domain_name == request["user"]["domain_name"]
    else:
        return request["user"]["uyud"] == resource_owner


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key("name", default=None): t.Null | t.String,
    })
)
async def list_serve(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]

    log.info("SERVE.LIST (email:{}, ak:{})", request["user"]["email"], access_key)
    query_conds = (EndpointRow.session_owner == request["user"]["uuid"]) & (
        EndpointRow.lifecycle_stage == EndpointLifecycle.CREATED
    )
    if params["name"]:
        query_conds &= EndpointRow.name == params["name"]

    async with root_ctx.db.begin_readonly_session() as db_sess:
        query = (
            sa.select(EndpointRow).where(query_conds).options(selectinload(EndpointRow.routings))
        )
        result = await db_sess.execute(query)
        rows = result.scalars().all()

    return web.json_response(
        [
            {
                "id": str(endpoint.id),
                "name": endpoint.name,
                "desired_session_count": endpoint.desired_session_count,
                "active_route_count": len([
                    r for r in endpoint.routings if r.status == RouteStatus.HEALTHY
                ]),
                "service_endpoint": endpoint.url,
                "is_public": endpoint.open_to_public,
            }
            for endpoint in rows
        ],
        status=200,
    )


@auth_required
@server_status_required(READ_ALLOWED)
async def get_info(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.GET_INFO (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound

    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})
    resp = {
        "endpoint_id": str(endpoint.id),
        "name": endpoint.name,
        "desired_session_count": endpoint.desired_session_count,
        "active_routes": [
            {
                "route_id": str(r.id),
                "session_id": str(r.session),
                "traffic_ratio": r.traffic_ratio,
            }
            for r in endpoint.routings
            if r.status == RouteStatus.HEALTHY
        ],
        "service_endpoint": endpoint.url,
        "is_public": endpoint.open_to_public,
    }

    return web.json_response(resp, status=200)


@auth_required
@server_status_required(ALL_ALLOWED)
@check_api_params(
    t.Dict({
        tx.AliasedKey(["name", "service_name", "clientSessionToken"]) >> "service_name": t.Regexp(
            r"^(?=.{4,24}$)\w[\w.-]*\w$", re.ASCII
        ),
        tx.AliasedKey(["desired_session_count", "desiredSessionCount"]): t.Int,
        tx.AliasedKey(["image", "lang"]): t.String,
        tx.AliasedKey(["arch", "architecture"], default=DEFAULT_IMAGE_ARCH)
        >> "architecture": t.String,
        tx.AliasedKey(["group", "groupName", "group_name"], default="default"): t.String,
        tx.AliasedKey(["domain", "domainName", "domain_name"], default="default"): t.String,
        tx.AliasedKey(["cluster_size", "clusterSize"], default=1): t.ToInt[1:],  # new in APIv6
        tx.AliasedKey(["cluster_mode", "clusterMode"], default="single-node"): tx.Enum(
            ClusterMode
        ),  # new in APIv6
        t.Key("tag", default=None): t.Null | t.String,
        tx.AliasedKey(["startup_command", "startupCommand"], default=None): t.Null | t.String,
        tx.AliasedKey(["bootstrap_script", "bootstrapScript"], default=None): t.Null | t.String,
        tx.AliasedKey(["callback_url", "callbackUrl", "callbackURL"], default=None): (
            t.Null | tx.URL
        ),
        t.Key("owner_access_key", default=None): t.Null | t.String,
        t.Key("open_to_public", default=False): t.Bool,
        t.Key("config"): t.Dict({
            t.Key("model"): t.String,
            tx.AliasedKey(["model_version", "modelVersion"], default=None): t.Null | t.String,
            tx.AliasedKey(
                ["model_mount_destination", "modelMountDestination"], default="/models"
            ): t.String,
            t.Key("environ", default=None): t.Null | t.Mapping(t.String, t.String),
            # cluster_size is moved to the root-level parameters
            tx.AliasedKey(["scaling_group", "scalingGroup"], default=None): t.Null | t.String,
            t.Key("resources", default=None): t.Null | t.Mapping(t.String, t.Any),
            tx.AliasedKey(["resource_opts", "resourceOpts"], default=None): t.Null
            | t.Mapping(t.String, t.Any),
        }),
    }),
)
async def create(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    scopes_param = {
        "owner_access_key": (
            None if params["owner_access_key"] is undefined else params["owner_access_key"]
        ),
    }
    requester_access_key, owner_access_key = await get_access_key_scopes(request, scopes_param)
    if params["desired_session_count"] > (
        _m := request["user"]["resource_policy"]["max_session_count_per_model_session"]
    ):
        raise InvalidAPIParameters(f"Cannot spawn more than {_m} sessions for a single service")

    async with root_ctx.db.begin_readonly() as conn:
        checked_scaling_group = await check_scaling_group(
            conn,
            params["config"]["scaling_group"],
            SessionTypes.INFERENCE,
            owner_access_key,
            params["domain"],
            params["group"],
        )

        query = (
            sa.select([scaling_groups.c.wsproxy_addr, scaling_groups.c.wsproxy_api_token])
            .select_from(scaling_groups)
            .where((scaling_groups.c.name == checked_scaling_group))
        )

        result = await conn.execute(query)
        sgroup = result.first()
        wsproxy_addr = sgroup["wsproxy_addr"]
        if not wsproxy_addr:
            raise ServiceUnavailable("No coordinator configured for this resource group")

        if not sgroup["wsproxy_api_token"]:
            raise ServiceUnavailable("Scaling group not ready to start model service")

        params["config"]["scaling_group"] = checked_scaling_group

        owner_uuid, group_id, resource_policy = await query_userinfo(request, params, conn)
        allowed_vfolder_types = await root_ctx.shared_config.get_vfolder_types()
        try:
            extra_vf_conds = vfolders.c.id == uuid.UUID(params["config"]["model"])
            matched_vfolders = await query_accessible_vfolders(
                conn,
                owner_uuid,
                user_role=request["user"]["role"],
                domain_name=params["domain"],
                allowed_vfolder_types=allowed_vfolder_types,
                extra_vf_conds=extra_vf_conds,
            )
        except Exception as e:
            # just catching ValueError | VFolderNotFound will raise
            # TypeError: catching classes that do not inherit from BaseException is not allowed
            if isinstance(e, ValueError) or isinstance(e, VFolderNotFound):
                try:
                    extra_vf_conds = (vfolders.c.name == params["config"]["model"]) & (
                        vfolders.c.usage_mode == VFolderUsageMode.MODEL
                    )
                    matched_vfolders = await query_accessible_vfolders(
                        conn,
                        owner_uuid,
                        user_role=request["user"]["role"],
                        domain_name=params["domain"],
                        allowed_vfolder_types=allowed_vfolder_types,
                        extra_vf_conds=extra_vf_conds,
                    )
                except VFolderNotFound as e:
                    raise VFolderNotFound("Cannot find model folder") from e
            else:
                raise
        if len(matched_vfolders) == 0:
            raise VFolderNotFound
        folder_row = matched_vfolders[0]
        if folder_row["usage_mode"] != VFolderUsageMode.MODEL:
            raise InvalidAPIParameters("Selected vFolder is not a model folder")

        model_id = folder_row["id"]

    proxy_name, volume_name = root_ctx.storage_manager.split_host(folder_row["host"])

    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/list",
        json={
            "volume": volume_name,
            "vfid": str(VFolderID(folder_row["quota_scope_id"], folder_row["id"])),
            "relpath": ".",
        },
    ) as (client_api_url, storage_resp):
        storage_reply = await storage_resp.json()

    for item in storage_reply["items"]:
        if item["name"] == "model-definition.yml" or item["name"] == "model-definition.yaml":
            yaml_name = item["name"]
            break
    else:
        raise InvalidAPIParameters("Model definition YAML file not found inside the model storage")

    chunks = bytes()
    async with root_ctx.storage_manager.request(
        proxy_name,
        "POST",
        "folder/file/fetch",
        json={
            "volume": volume_name,
            "vfid": str(VFolderID(folder_row["quota_scope_id"], folder_row["id"])),
            "relpath": f"./{yaml_name}",
        },
    ) as (client_api_url, storage_resp):
        while True:
            chunk = await storage_resp.content.read(DEFAULT_CHUNK_SIZE)
            if not chunk:
                break
            chunks += chunk
    model_definition_yaml = chunks.decode("utf-8")
    model_definition_dict = yaml.load(model_definition_yaml, Loader=yaml.FullLoader)
    try:
        model_definition = model_definition_iv.check(model_definition_dict)
        assert model_definition is not None
    except t.DataError as e:
        raise InvalidAPIParameters(
            f"Failed to validate model definition from vFolder {folder_row['name']} (ID"
            f" {folder_row['id']}): {e}",
        ) from e
    except yaml.error.YAMLError as e:
        raise InvalidAPIParameters(f"Invalid YAML syntax: {e}") from e

    async with root_ctx.db.begin_readonly_session() as session:
        image_row = await ImageRow.resolve(
            session,
            [
                ImageRef(params["image"], ["*"], params["architecture"]),
                params["image"],
            ],
        )

    params["config"]["mount_map"] = {model_id: params["config"]["model_mount_destination"]}
    sudo_session_enabled = request["user"]["sudo_session_enabled"]

    # check if session is valid to be created
    await root_ctx.registry.create_session(
        "",
        params["image"],
        params["architecture"],
        UserScope(
            domain_name=params["domain"],
            group_id=group_id,
            user_uuid=request["user"]["uuid"],
            user_role=request["user"]["role"],
        ),
        owner_access_key,
        resource_policy,
        SessionTypes.INFERENCE,
        params["config"],
        params["cluster_mode"],
        params["cluster_size"],
        dry_run=True,  # Setting this to True will prevent actual session from being enqueued
        bootstrap_script=params["bootstrap_script"],
        startup_command=params["startup_command"],
        tag=params["tag"],
        callback_url=params["callback_url"],
        sudo_session_enabled=sudo_session_enabled,
    )

    async with root_ctx.db.begin_session() as db_sess:
        query = sa.select(EndpointRow).where(
            (EndpointRow.lifecycle_stage != EndpointLifecycle.DESTROYED)
            & (EndpointRow.name == params["service_name"])
        )
        result = await db_sess.execute(query)
        service_with_duplicate_name = result.scalar()
        if service_with_duplicate_name is not None:
            raise InvalidAPIParameters("Cannot create multiple services with same name")

        project_id = await resolve_group_name_or_id(
            await db_sess.connection(), params["domain"], params["group"]
        )
        if project_id is None:
            raise InvalidAPIParameters(f"Invalid group name {project_id}")
        endpoint = EndpointRow(
            params["service_name"],
            request["user"]["uuid"],
            owner_uuid,
            params["desired_session_count"],
            image_row,
            model_id,
            params["domain"],
            project_id,
            checked_scaling_group,
            params["config"]["resources"],
            params["cluster_mode"],
            params["cluster_size"],
            model_mount_destination=params["config"]["model_mount_destination"],
            tag=params["tag"],
            startup_command=params["startup_command"],
            callback_url=params["callback_url"],
            environ=params["config"]["environ"],
            bootstrap_script=params["bootstrap_script"],
            resource_opts=params["config"]["resource_opts"],
            open_to_public=params["open_to_public"],
        )
        db_sess.add(endpoint)
        await db_sess.commit()

    return web.json_response({"endpoint_id": str(endpoint.id)})


@auth_required
@server_status_required(READ_ALLOWED)
async def delete(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.DELETE (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    async with root_ctx.db.begin_session() as db_sess:
        if len(endpoint.routings) == 0:
            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == service_id)
                .values({"lifecycle_stage": EndpointLifecycle.DESTROYED})
            )
        else:
            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == service_id)
                .values({
                    "desired_session_count": 0,
                    "lifecycle_stage": EndpointLifecycle.DESTROYING,
                })
            )
        await db_sess.execute(query)
    return web.json_response({"success": True}, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
async def sync(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info("SERVE.SYNC (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id)

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    async with root_ctx.db.begin_session() as db_sess:
        await root_ctx.registry.update_appproxy_endpoint_routes(
            db_sess, endpoint, [r for r in endpoint.routings if r.status == RouteStatus.HEALTHY]
        )
    return web.json_response({"success": True}, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key("to"): t.Int,
    }),
)
async def scale(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.SCALE (email:{}, ak:{}, s:{})", request["user"]["email"], access_key, service_id
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    if params["to"] < 0:
        raise InvalidAPIParameters("Amount of desired session count cannot be a negative number")
    elif params["to"] > (
        _m := request["user"]["resource_policy"]["max_session_count_per_model_session"]
    ):
        raise InvalidAPIParameters(f"Cannot spawn more than {_m} sessions for a single service")

    async with root_ctx.db.begin_session() as db_sess:
        query = (
            sa.update(EndpointRow)
            .where(EndpointRow.id == service_id)
            .values({"desired_session_count": params["to"]})
        )
        await db_sess.execute(query)
        return web.json_response({
            "current_route_count": len(endpoint.routings),
            "target_count": params["to"],
        })


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key("traffic_ratio"): t.Float[0:],
    }),
)
async def update_route(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])
    route_id = uuid.UUID(request.match_info["route_id"])

    log.info(
        "SERVE.UPDATE_ROUTE (email:{}, ak:{}, s:{}, r:{})",
        request["user"]["email"],
        access_key,
        service_id,
        route_id,
    )

    async with root_ctx.db.begin_session() as db_sess:
        try:
            route = await RoutingRow.get(db_sess, route_id, load_endpoint=True)
        except NoResultFound:
            raise ObjectNotFound
        if route.endpoint != service_id:
            raise ObjectNotFound
        await get_user_uuid_scopes(request, {"owner_uuid": route.endpoint_row.session_owner})

        query = (
            sa.update(RoutingRow)
            .where(RoutingRow.id == route_id)
            .values({"traffic_ratio": params["traffic_ratio"]})
        )
        await db_sess.execute(query)
        endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        try:
            await root_ctx.registry.update_appproxy_endpoint_routes(
                db_sess, endpoint, [r for r in endpoint.routes if r.status == RouteStatus.HEALTHY]
            )
        except aiohttp.ClientError as e:
            log.warning("failed to communicate with AppProxy endpoint: {}", str(e))
        return web.json_response({"success": True})


@auth_required
@server_status_required(READ_ALLOWED)
async def delete_route(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])
    route_id = uuid.UUID(request.match_info["route_id"])

    log.info(
        "SERVE.DELETE_ROUTE (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )
    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            route = await RoutingRow.get(db_sess, route_id, load_session=True)
        except NoResultFound:
            raise ObjectNotFound
        if route.endpoint != service_id:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": route.endpoint_row.session_owner})
    if route.status == RouteStatus.PROVISIONING:
        raise InvalidAPIParameters("Cannot remove route in PROVISIONING status")

    await root_ctx.registry.destroy_session(
        route.session_row,
        forced=False,
        reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
    )

    async with root_ctx.db.begin_session() as db_sess:
        query = (
            sa.update(EndpointRow)
            .where(EndpointRow.id == service_id)
            .values({"desired_session_count": route.endpoint_row.desired_session_count})
        )
        await db_sess.execute(query)
        return web.json_response({"success": True})


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key("duration", default=None): t.Null | tx.TimeDuration,
        t.Key("valid_until", default=None): t.Null | t.Int,
    }),
)
async def generate_token(request: web.Request, params: Any) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.GENERATE_TOKEN (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
        query = (
            sa.select([scaling_groups.c.wsproxy_addr, scaling_groups.c.wsproxy_api_token])
            .select_from(scaling_groups)
            .where((scaling_groups.c.name == endpoint.resource_group))
        )

        result = await db_sess.execute(query)
        sgroup = result.first()
        wsproxy_addr = sgroup["wsproxy_addr"]
        wsproxy_api_token = sgroup["wsproxy_api_token"]

    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    if params["valid_until"]:
        exp = params["valid_until"]
    elif params["duration"]:
        exp = int((datetime.now() + params["duration"]).timestamp())
    else:
        raise InvalidAPIParameters("valid_until and duration can't be both unspecified")
    if datetime.now().timestamp() > exp:
        raise InvalidAPIParameters("valid_until is older than now")
    body = {"user_uuid": str(endpoint.session_owner), "exp": exp}
    async with aiohttp.ClientSession() as session:
        async with session.post(
            f"{wsproxy_addr}/v2/endpoints/{endpoint.id}/token",
            json=body,
            headers={
                "X-BackendAI-Token": wsproxy_api_token,
            },
        ) as resp:
            token_json = await resp.json()
            token = token_json["token"]

    async with root_ctx.db.begin_session() as db_sess:
        token_row = EndpointTokenRow(
            uuid.uuid4(),
            token,
            endpoint.id,
            endpoint.domain,
            endpoint.project,
            endpoint.session_owner,
        )
        db_sess.add(token_row)
        await db_sess.commit()
        return web.json_response({"token": token})


@auth_required
@server_status_required(READ_ALLOWED)
async def list_errors(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.LIST_ERRORS (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    error_routes = [r for r in endpoint.routings if r.status == RouteStatus.FAILED_TO_START]

    return web.json_response({
        "errors": [
            {
                "session_id": route.error_data.get("session_id"),
                "error": route.error_data["errors"],
            }
            for route in error_routes
        ],
        "retries": endpoint.retries,
    })


@auth_required
@server_status_required(READ_ALLOWED)
async def clear_error(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    access_key = request["keypair"]["access_key"]
    service_id = uuid.UUID(request.match_info["service_id"])

    log.info(
        "SERVE.CLEAR_ERROR (email:{}, ak:{}, s:{})",
        request["user"]["email"],
        access_key,
        service_id,
    )

    async with root_ctx.db.begin_readonly_session() as db_sess:
        try:
            endpoint = await EndpointRow.get(db_sess, service_id, load_routes=True)
        except NoResultFound:
            raise ObjectNotFound
    await get_user_uuid_scopes(request, {"owner_uuid": endpoint.session_owner})

    async with root_ctx.db.begin_session() as db_sess:
        query = sa.delete(RoutingRow).where(
            (RoutingRow.endpoint == service_id) & (RoutingRow.status == RouteStatus.FAILED_TO_START)
        )
        await db_sess.execute(query)
        query = sa.update(EndpointRow).values({"retries": 0}).where(EndpointRow.id == endpoint.id)
        await db_sess.execute(query)

    return web.Response(status=204)


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    database_ptask_group: aiotools.PersistentTaskGroup


async def init(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    app_ctx.database_ptask_group = aiotools.PersistentTaskGroup()


async def shutdown(app: web.Application) -> None:
    app_ctx: PrivateContext = app["services.context"]
    await app_ctx.database_ptask_group.shutdown()


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "services"
    app["api_versions"] = (4, 5)
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["services.context"] = PrivateContext()
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", list_serve))
    cors.add(root_resource.add_route("POST", create))
    cors.add(add_route("GET", "/{service_id}", get_info))
    cors.add(add_route("DELETE", "/{service_id}", delete))
    cors.add(add_route("GET", "/{service_id}/errors", list_errors))
    cors.add(add_route("POST", "/{service_id}/errors/clear", clear_error))
    cors.add(add_route("POST", "/{service_id}/scale", scale))
    cors.add(add_route("POST", "/{service_id}/sync", sync))
    cors.add(add_route("PUT", "/{service_id}/routings/{route_id}", update_route))
    cors.add(add_route("DELETE", "/{service_id}/routings/{route_id}", delete_route))
    cors.add(add_route("POST", "/{service_id}/token", generate_token))
    return app, []



from __future__ import annotations

import asyncio
import json
import logging
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager as actxmgr
from contextvars import ContextVar
from typing import Any, AsyncIterator, Dict, Final, Mapping, Optional, cast

import aiohttp
import aiotools
import sqlalchemy as sa
import trafaret as t
import yarl

from ai.backend.common.bgtask import ProgressReporter
from ai.backend.common.docker import ImageRef, arch_name_aliases, validate_image_labels
from ai.backend.common.docker import login as registry_login
from ai.backend.common.exception import InvalidImageName, InvalidImageTag
from ai.backend.common.logging import BraceStyleAdapter

from ..models.image import ImageRow, ImageType
from ..models.utils import ExtendedAsyncSAEngine

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]
concurrency_sema: ContextVar[asyncio.Semaphore] = ContextVar("concurrency_sema")
progress_reporter: ContextVar[Optional[ProgressReporter]] = ContextVar(
    "progress_reporter", default=None
)
all_updates: ContextVar[Dict[ImageRef, Dict[str, Any]]] = ContextVar("all_updates")


class BaseContainerRegistry(metaclass=ABCMeta):
    db: ExtendedAsyncSAEngine
    registry_name: str
    registry_info: Mapping[str, Any]
    registry_url: yarl.URL
    max_concurrency_per_registry: int
    base_hdrs: Dict[str, str]
    credentials: Dict[str, str]
    ssl_verify: bool

    MEDIA_TYPE_OCI_INDEX: Final[str] = "application/vnd.oci.image.index.v1+json"
    MEDIA_TYPE_OCI_MANIFEST: Final[str] = "application/vnd.oci.image.manifest.v1+json"
    MEDIA_TYPE_DOCKER_MANIFEST_LIST: Final[str] = (
        "application/vnd.docker.distribution.manifest.list.v2+json"
    )
    MEDIA_TYPE_DOCKER_MANIFEST: Final[str] = "application/vnd.docker.distribution.manifest.v2+json"

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        registry_name: str,
        registry_info: Mapping[str, Any],
        *,
        max_concurrency_per_registry: int = 4,
        ssl_verify: bool = True,
    ) -> None:
        self.db = db
        self.registry_name = registry_name
        self.registry_info = registry_info
        self.registry_url = registry_info[""]
        self.max_concurrency_per_registry = max_concurrency_per_registry
        self.base_hdrs = {
            "Accept": "application/vnd.docker.distribution.manifest.v2+json",
        }
        self.credentials = {}
        self.ssl_verify = ssl_verify

    @actxmgr
    async def prepare_client_session(self) -> AsyncIterator[tuple[yarl.URL, aiohttp.ClientSession]]:
        ssl_ctx = None  # default
        if not self.registry_info["ssl_verify"]:
            ssl_ctx = False
        connector = aiohttp.TCPConnector(ssl=ssl_ctx)
        async with aiohttp.ClientSession(connector=connector) as sess:
            yield self.registry_url, sess

    async def rescan_single_registry(
        self,
        reporter: ProgressReporter | None = None,
    ) -> None:
        all_updates_token = all_updates.set({})
        concurrency_sema.set(asyncio.Semaphore(self.max_concurrency_per_registry))
        progress_reporter.set(reporter)
        try:
            username = self.registry_info["username"]
            if username is not None:
                self.credentials["username"] = username
            password = self.registry_info["password"]
            if password is not None:
                self.credentials["password"] = password
            async with self.prepare_client_session() as (url, client_session):
                self.registry_url = url
                async with aiotools.TaskGroup() as tg:
                    async for image in self.fetch_repositories(client_session):
                        tg.create_task(self._scan_image(client_session, image))
            await self.commit_rescan_result()
        finally:
            all_updates.reset(all_updates_token)

    async def commit_rescan_result(self) -> None:
        _all_updates = all_updates.get()
        if not _all_updates:
            log.info("No images found in registry {0}", self.registry_url)
        else:
            image_identifiers = [(k.canonical, k.architecture) for k in _all_updates.keys()]
            async with self.db.begin_session() as session:
                existing_images = await session.scalars(
                    sa.select(ImageRow).where(
                        sa.func.ROW(ImageRow.name, ImageRow.architecture).in_(image_identifiers),
                    ),
                )
                is_local = self.registry_name == "local"

                for image_row in existing_images:
                    key = image_row.image_ref
                    values = _all_updates.get(key)
                    if values is None:
                        continue
                    _all_updates.pop(key)
                    image_row.config_digest = values["config_digest"]
                    image_row.size_bytes = values["size_bytes"]
                    image_row.accelerators = values.get("accels")
                    image_row.labels = values["labels"]
                    image_row.is_local = is_local
                    image_row.resources = values["resources"]

                session.add_all([
                    ImageRow(
                        name=k.canonical,
                        registry=k.registry,
                        image=k.name,
                        tag=k.tag,
                        architecture=k.architecture,
                        is_local=is_local,
                        config_digest=v["config_digest"],
                        size_bytes=v["size_bytes"],
                        type=ImageType.COMPUTE,
                        accelerators=v.get("accels"),
                        labels=v["labels"],
                        resources=v["resources"],
                    )
                    for k, v in _all_updates.items()
                ])
                await session.flush()

    async def scan_single_ref(self, image_ref: str) -> None:
        all_updates_token = all_updates.set({})
        sema_token = concurrency_sema.set(asyncio.Semaphore(1))
        try:
            async with self.prepare_client_session() as (url, sess):
                image, tag = ImageRef._parse_image_tag(image_ref)
                rqst_args = await registry_login(
                    sess,
                    self.registry_url,
                    self.credentials,
                    f"repository:{image}:pull",
                )
                rqst_args["headers"].update(**self.base_hdrs)
                await self._scan_tag(sess, rqst_args, image, tag)
            await self.commit_rescan_result()
        finally:
            concurrency_sema.reset(sema_token)
            all_updates.reset(all_updates_token)

    async def _scan_image(
        self,
        sess: aiohttp.ClientSession,
        image: str,
    ) -> None:
        rqst_args = await registry_login(
            sess,
            self.registry_url,
            self.credentials,
            f"repository:{image}:pull",
        )
        rqst_args["headers"].update(**self.base_hdrs)
        tags = []
        tag_list_url: Optional[yarl.URL]
        tag_list_url = (self.registry_url / f"v2/{image}/tags/list").with_query(
            {"n": "10"},
        )
        while tag_list_url is not None:
            async with sess.get(tag_list_url, **rqst_args) as resp:
                data = json.loads(await resp.read())
                if "tags" in data:
                    # sometimes there are dangling image names in the hub.
                    tags.extend(data["tags"])
                tag_list_url = None
                next_page_link = resp.links.get("next")
                if next_page_link:
                    next_page_url = cast(yarl.URL, next_page_link["url"])
                    tag_list_url = self.registry_url.with_path(next_page_url.path).with_query(
                        next_page_url.query
                    )
        if (reporter := progress_reporter.get()) is not None:
            reporter.total_progress += len(tags)
        async with aiotools.TaskGroup() as tg:
            for tag in tags:
                tg.create_task(self._scan_tag(sess, rqst_args, image, tag))

    async def _scan_tag(
        self,
        sess: aiohttp.ClientSession,
        rqst_args: dict[str, Any],
        image: str,
        tag: str,
    ) -> None:
        manifests = {}
        async with concurrency_sema.get():
            rqst_args["headers"]["Accept"] = self.MEDIA_TYPE_DOCKER_MANIFEST_LIST
            async with sess.get(
                self.registry_url / f"v2/{image}/manifests/{tag}", **rqst_args
            ) as resp:
                if resp.status == 404:
                    # ignore missing tags
                    # (may occur after deleting an image from the docker hub)
                    return
                content_type = resp.headers["Content-Type"]
                resp.raise_for_status()
                resp_json = await resp.json()
                match content_type:
                    case self.MEDIA_TYPE_DOCKER_MANIFEST_LIST:
                        manifest_list = resp_json["manifests"]
                        request_type = self.MEDIA_TYPE_DOCKER_MANIFEST
                    case self.MEDIA_TYPE_OCI_INDEX:
                        manifest_list = [
                            item
                            for item in resp_json["manifests"]
                            if "annotations" not in item  # skip attestation manifests
                        ]
                        request_type = self.MEDIA_TYPE_OCI_MANIFEST
                    case _:
                        raise RuntimeError(
                            "The registry does not support the standard way of "
                            "listing multiarch images."
                        )
            rqst_args["headers"]["Accept"] = request_type
            for manifest in manifest_list:
                platform_arg = (
                    f"{manifest['platform']['os']}/{manifest['platform']['architecture']}"
                )
                if variant := manifest["platform"].get("variant", None):
                    platform_arg += f"/{variant}"
                architecture = manifest["platform"]["architecture"]
                architecture = arch_name_aliases.get(architecture, architecture)
                async with sess.get(
                    self.registry_url / f"v2/{image}/manifests/{manifest['digest']}", **rqst_args
                ) as resp:
                    data = await resp.json()
                config_digest = data["config"]["digest"]
                size_bytes = sum(layer["size"] for layer in data["layers"]) + data["config"]["size"]
                async with sess.get(
                    self.registry_url / f"v2/{image}/blobs/{config_digest}", **rqst_args
                ) as resp:
                    resp.raise_for_status()
                    data = json.loads(await resp.read())
                labels = {}
                if "container_config" in data:
                    raw_labels = data["container_config"].get("Labels")
                    if raw_labels:
                        labels.update(raw_labels)
                    else:
                        log.warning(
                            "label not found on image {}:{}/{}",
                            image,
                            tag,
                            architecture,
                        )
                else:
                    raw_labels = data["config"].get("Labels")
                    if raw_labels:
                        labels.update(raw_labels)
                    else:
                        log.warning(
                            "label not found on image {}:{}/{}",
                            image,
                            tag,
                            architecture,
                        )
                manifests[architecture] = {
                    "size": size_bytes,
                    "labels": labels,
                    "digest": config_digest,
                }
            await self._read_manifest(image, tag, manifests)

    async def _read_manifest(
        self,
        image: str,
        tag: str,
        manifests: dict[str, dict],
        skip_reason: Optional[str] = None,
    ) -> None:
        if not manifests:
            if not skip_reason:
                skip_reason = "missing/deleted"
            log.warning("Skipped image - {}:{} ({})", image, tag, skip_reason)
            progress_msg = f"Skipped {image}:{tag} ({skip_reason})"
            if (reporter := progress_reporter.get()) is not None:
                await reporter.update(1, message=progress_msg)
            return

        assert ImageRow.resources is not None
        for architecture, manifest in manifests.items():
            try:
                try:
                    validate_image_labels(manifest["labels"])
                except t.DataError as e:
                    match e.as_dict():
                        case str() as error_msg:
                            skip_reason = error_msg
                        case dict() as error_data:
                            skip_reason = "; ".join(
                                f"{field} {reason}" for field, reason in error_data.items()
                            )
                    continue
                except ValueError as e:
                    skip_reason = str(e)
                    continue
                if self.registry_name == "local":
                    if image.partition("/")[1] == "":
                        image = "library/" + image
                    update_key = ImageRef(
                        f"{image}:{tag}",
                        ["index.docker.io"],
                        architecture,
                    )
                else:
                    update_key = ImageRef(
                        f"{self.registry_name}/{image}:{tag}",
                        [self.registry_name],
                        architecture,
                    )
                updates = {
                    "config_digest": manifest["digest"],
                    "size_bytes": manifest["size"],
                    "labels": manifest["labels"],  # keep the original form
                }
                accels = manifest["labels"].get("ai.backend.accelerators")
                if accels:
                    updates["accels"] = accels

                resources = {  # default fallback if not defined
                    "cpu": {"min": "1", "max": None},
                    "mem": {"min": "1g", "max": None},
                }
                res_prefix = "ai.backend.resource.min."
                for k, v in filter(
                    lambda pair: pair[0].startswith(res_prefix), manifest["labels"].items()
                ):
                    res_key = k[len(res_prefix) :]
                    resources[res_key] = {"min": v}
                updates["resources"] = ImageRow.resources.type._schema.check(resources)
                all_updates.get().update({
                    update_key: updates,
                })
            except (InvalidImageName, InvalidImageTag) as e:
                skip_reason = str(e)
            finally:
                if skip_reason:
                    log.warning(
                        "Skipped image - {}:{}/{} ({})", image, tag, architecture, skip_reason
                    )
                    progress_msg = f"Skipped {image}:{tag}/{architecture} ({skip_reason})"
                else:
                    log.info(
                        "Updated image - {0}:{1}/{2} ({3})",
                        image,
                        tag,
                        architecture,
                        manifest["digest"],
                    )
                    progress_msg = f"Updated {image}:{tag}/{architecture} ({manifest['digest']})"
                if (reporter := progress_reporter.get()) is not None:
                    await reporter.update(1, message=progress_msg)

    @abstractmethod
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        yield ""



from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext


class ManagerErrorPluginContext(ErrorPluginContext):
    blocklist = {"ai.backend.agent"}


class ManagerStatsPluginContext(StatsPluginContext):
    blocklist = {"ai.backend.agent"}



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Mapping, Sequence

import graphene

from ai.backend.common.logging import BraceStyleAdapter

from ..defs import PASSWORD_PLACEHOLDER
from . import UserRole
from .base import privileged_mutation, set_if_set

if TYPE_CHECKING:
    from .gql import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.models.etcd"))  # type: ignore[name-defined]

__all__: Sequence[str] = (
    "ContainerRegistry",
    "CreateContainerRegistry",
    "ModifyContainerRegistry",
    "DeleteContainerRegistry",
)


class CreateContainerRegistryInput(graphene.InputObjectType):
    url = graphene.String(required=True)
    type = graphene.String(required=True)
    project = graphene.List(graphene.String)
    username = graphene.String()
    password = graphene.String()
    ssl_verify = graphene.Boolean()


class ModifyContainerRegistryInput(graphene.InputObjectType):
    url = graphene.String()
    type = graphene.String()
    project = graphene.List(graphene.String)
    username = graphene.String()
    password = graphene.String()
    ssl_verify = graphene.Boolean()


class ContainerRegistryConfig(graphene.ObjectType):
    url = graphene.String(required=True)
    type = graphene.String(required=True)
    project = graphene.List(graphene.String)
    username = graphene.String()
    password = graphene.String()
    ssl_verify = graphene.Boolean()


class ContainerRegistry(graphene.ObjectType):
    hostname = graphene.String()
    config = graphene.Field(ContainerRegistryConfig)

    class Meta:
        interfaces = (graphene.relay.Node,)

    # TODO: `get_node()` should be implemented to query a scalar object directly by ID
    #       (https://docs.graphene-python.org/en/latest/relay/nodes/#nodes)
    # @classmethod
    # def get_node(cls, info: graphene.ResolveInfo, id):
    #     raise NotImplementedError

    @classmethod
    def from_row(cls, hostname: str, config: Mapping[str, str | list | None]) -> ContainerRegistry:
        password = config.get("password", None)
        return cls(
            id=hostname,
            hostname=hostname,
            config=ContainerRegistryConfig(
                url=config.get(""),
                type=config.get("type"),
                project=config.get("project", None),
                username=config.get("username", None),
                password=PASSWORD_PLACEHOLDER if password is not None else None,
                ssl_verify=config.get("ssl_verify", None),
            ),
        )

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
    ) -> Sequence[ContainerRegistry]:
        log.info(
            "ETCD.LIST_CONTAINER_REGISTRY (ak:{})",
            ctx.access_key,
        )
        registries = await ctx.shared_config.list_container_registry()
        return [cls.from_row(hostname, config) for hostname, config in registries.items()]

    @classmethod
    async def load_registry(cls, ctx: GraphQueryContext, hostname: str) -> ContainerRegistry:
        log.info(
            "ETCD.GET_CONTAINER_REGISTRY (ak:{}, hostname:{})",
            ctx.access_key,
            hostname,
        )
        item = await ctx.shared_config.get_container_registry(hostname)
        return cls.from_row(hostname, item)


class CreateContainerRegistry(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistry)

    class Arguments:
        hostname = graphene.String(required=True)
        props = CreateContainerRegistryInput(required=True)

    @classmethod
    @privileged_mutation(
        UserRole.SUPERADMIN,
        lambda id, **kwargs: (None, id),
    )
    async def mutate(
        cls, root, info: graphene.ResolveInfo, hostname: str, props: CreateContainerRegistryInput
    ) -> CreateContainerRegistry:
        ctx: GraphQueryContext = info.context
        input_config: Dict[str, Any] = {"": props.url, "type": props.type}
        set_if_set(props, input_config, "project")
        set_if_set(props, input_config, "username")
        set_if_set(props, input_config, "password")
        set_if_set(props, input_config, "ssl_verify")
        log.info(
            "ETCD.CREATE_CONTAINER_REGISTRY (ak:{}, hostname:{}, config:{})",
            ctx.access_key,
            hostname,
            input_config,
        )
        await ctx.shared_config.add_container_registry(hostname, input_config)
        container_registry = await ContainerRegistry.load_registry(ctx, hostname)
        return cls(container_registry=container_registry)


class ModifyContainerRegistry(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistry)

    class Arguments:
        hostname = graphene.String(required=True)
        props = ModifyContainerRegistryInput(required=True)

    @classmethod
    @privileged_mutation(
        UserRole.SUPERADMIN,
        lambda id, **kwargs: (None, id),
    )
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        hostname: str,
        props: ModifyContainerRegistryInput,
    ) -> ModifyContainerRegistry:
        ctx: GraphQueryContext = info.context
        input_config: Dict[str, Any] = {}
        set_if_set(props, input_config, "url")
        set_if_set(props, input_config, "type")
        set_if_set(props, input_config, "project")
        set_if_set(props, input_config, "username")
        set_if_set(props, input_config, "password")
        set_if_set(props, input_config, "ssl_verify")
        if "url" in input_config:
            input_config[""] = input_config.pop("url")
        log.info(
            "ETCD.MODIFY_CONTAINER_REGISTRY (ak:{}, hostname:{}, config:{})",
            ctx.access_key,
            hostname,
            input_config,
        )
        await ctx.shared_config.modify_container_registry(hostname, input_config)
        container_registry = await ContainerRegistry.load_registry(ctx, hostname)
        return cls(container_registry=container_registry)


class DeleteContainerRegistry(graphene.Mutation):
    allowed_roles = (UserRole.SUPERADMIN,)
    container_registry = graphene.Field(ContainerRegistry)

    class Arguments:
        hostname = graphene.String(required=True)

    @classmethod
    @privileged_mutation(
        UserRole.SUPERADMIN,
        lambda id, **kwargs: (None, id),
    )
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        hostname: str,
    ) -> DeleteContainerRegistry:
        ctx: GraphQueryContext = info.context
        log.info(
            "ETCD.DELETE_CONTAINER_REGISTRY (ak:{}, hostname:{})",
            ctx.access_key,
            hostname,
        )
        container_registry = await ContainerRegistry.load_registry(ctx, hostname)
        await ctx.shared_config.delete_container_registry(hostname)
        return cls(container_registry=container_registry)



"""add-scheduled-to-kernelstatus

Revision ID: 8679d0a7e22b
Revises: 518ecf41f567
Create Date: 2021-04-01 14:24:27.885209

"""

import textwrap

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "8679d0a7e22b"
down_revision = "518ecf41f567"
branch_labels = None
depends_on = None

kernelstatus_new_values = [
    "PENDING",
    "SCHEDULED",  # added
    "PREPARING",
    "BUILDING",
    "PULLING",
    "RUNNING",
    "RESTARTING",
    "RESIZING",
    "SUSPENDED",
    "TERMINATING",
    "TERMINATED",
    "ERROR",
    "CANCELLED",
]
kernelstatus_new = postgresql.ENUM(*kernelstatus_new_values, name="kernelstatus")

kernelstatus_old_values = [
    "PENDING",
    "PREPARING",
    "BUILDING",
    "PULLING",
    "RUNNING",
    "RESTARTING",
    "RESIZING",
    "SUSPENDED",
    "TERMINATING",
    "TERMINATED",
    "ERROR",
    "CANCELLED",
]
kernelstatus_old = postgresql.ENUM(*kernelstatus_old_values, name="kernelstatus")


def upgrade():
    conn = op.get_bind()
    conn.execute(text("DROP INDEX IF EXISTS ix_kernels_unique_sess_token;"))
    conn.execute(text("ALTER TYPE kernelstatus RENAME TO kernelstatus_old;"))
    kernelstatus_new.create(conn)
    conn.execute(
        text(
            textwrap.dedent(
                """\
            ALTER TABLE kernels
                ALTER COLUMN "status" DROP DEFAULT,
                ALTER COLUMN "status" TYPE kernelstatus USING "status"::text::kernelstatus,
                ALTER COLUMN "status" SET DEFAULT 'PENDING'::kernelstatus;
            """
            )
        )
    )
    conn.execute(text("DROP TYPE kernelstatus_old;"))
    # This also fixes the unique constraint columns:
    #   (access_key, session_id) -> (access_key, session_name)
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "session_name"],
        unique=True,
        postgresql_where=sa.text(
            "status NOT IN ('TERMINATED', 'CANCELLED') and cluster_role = 'main'"
        ),
    )


def downgrade():
    op.drop_index("ix_kernels_unique_sess_token", table_name="kernels")
    conn = op.get_bind()
    conn.execute(text("ALTER TYPE kernelstatus RENAME TO kernelstatus_new;"))
    kernelstatus_old.create(conn)
    conn.execute(
        text(
            textwrap.dedent(
                """\
            ALTER TABLE kernels
                ALTER COLUMN "status" DROP DEFAULT,
                ALTER COLUMN "status" TYPE kernelstatus USING (
                    CASE "status"::text
                        WHEN 'SCHEDULED' THEN 'PREPARING'
                        ELSE "status"::text
                    END
                )::kernelstatus,
                ALTER COLUMN "status" SET DEFAULT 'PENDING'::kernelstatus;
            """
            )
        )
    )
    conn.execute(text("DROP TYPE kernelstatus_new;"))
    op.create_index(
        "ix_kernels_unique_sess_token",
        "kernels",
        ["access_key", "session_id"],
        unique=True,
        postgresql_where=sa.text("status != 'TERMINATED' and cluster_role = 'main'"),
    )



import enum
from typing import Mapping, NamedTuple, TypeAlias

import sqlalchemy as sa
from lark import Lark, LarkError, Transformer
from lark.lexer import Token

from . import JSONFieldItem, OrderSpecItem, get_col_from_table

__all__ = (
    "ColumnMapType",
    "QueryOrderParser",
)

_grammar = r"""
    ?start: expr
    expr   : [col ("," col)*]
    col    : ORDER? CNAME
    ORDER  : "+" | "-"
    %import common.CNAME
    %import common.WS
    %ignore WS
"""
_parser = Lark(
    _grammar,
    parser="lalr",
    maybe_placeholders=False,
)

ColumnMapType: TypeAlias = Mapping[str, OrderSpecItem] | None


class OrderDirection(enum.Enum):
    ASC = "asc"
    DESC = "desc"


class OrderingItem(NamedTuple):
    column: sa.Column
    order_direction: OrderDirection


class QueryOrderTransformer(Transformer):
    def __init__(self, sa_table: sa.Table, column_map: ColumnMapType = None) -> None:
        super().__init__()
        self._sa_table = sa_table
        self._column_map = column_map

    def _get_col(self, col_name: str) -> sa.Column:
        try:
            if self._column_map:
                col_value, func = self._column_map[col_name]
                match col_value:
                    case str(column):
                        matched_col = get_col_from_table(self._sa_table, column)
                    case JSONFieldItem(_col, _key):
                        _column = get_col_from_table(self._sa_table, _col)
                        matched_col = _column.op("->>")(_key)
                    case _:
                        raise ValueError("Invalid type of field name", col_name)
                col = func(matched_col) if func is not None else matched_col
            else:
                col = get_col_from_table(self._sa_table, col_name)
            return col
        except KeyError:
            raise ValueError("Unknown/unsupported field name", col_name)

    def col(self, *args) -> OrderingItem:
        children: list[Token] = args[0]
        if len(children) == 2:
            op = children[0].value
            col = self._get_col(children[1].value)
        else:
            op = "+"  # assume ascending if not marked
            col = self._get_col(children[0].value)
        if op == "+":
            return OrderingItem(col, OrderDirection.ASC)
        elif op == "-":
            return OrderingItem(col, OrderDirection.DESC)
        raise ValueError(f"Invalid operation `{op}`. Please use `+` or `-`")

    expr = tuple


class QueryOrderParser:
    def __init__(self, column_map: ColumnMapType = None) -> None:
        self._column_map = column_map
        self._parser = _parser

    def parse_order(self, table, order_expr: str) -> list[OrderingItem]:
        try:
            ast = self._parser.parse(order_expr)
            orders = QueryOrderTransformer(table, self._column_map).transform(ast)
            return orders
        except LarkError as e:
            raise ValueError(f"Query ordering parsing error: {e}")

    def append_ordering(
        self,
        sa_query: sa.sql.Select,
        order_expr: str,
    ) -> sa.sql.Select:
        """
        Parse the given filter expression and build the where clause based on the first target table from
        the given SQLAlchemy query object.
        """
        table = sa_query.froms[0]
        orders = [
            col.asc() if direction == OrderDirection.ASC else col.desc()
            for col, direction in self.parse_order(table, order_expr)
        ]
        return sa_query.order_by(*orders)



from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles
import click
import graphene

from ..models.gql import Mutations, Queries

if TYPE_CHECKING:
    from .context import CLIContext

log = logging.getLogger(__spec__.name)  # type: ignore[name-defined]


@click.group()
def cli(args) -> None:
    pass


async def generate_gql_schema(output_path: Path) -> None:
    schema = graphene.Schema(query=Queries, mutation=Mutations, auto_camelcase=False)
    if output_path == "-":
        log.info("======== GraphQL API Schema ========")
        print(str(schema))
    else:
        async with aiofiles.open(output_path, "w") as fw:
            await fw.write(str(schema))


@cli.command()
@click.pass_obj
@click.option(
    "--output",
    "-o",
    default="-",
    type=click.Path(dir_okay=False, writable=True),
    help="Output file path (default: stdout)",
)
def dump_gql_schema(cli_ctx: CLIContext, output: Path) -> None:
    asyncio.run(generate_gql_schema(output))



from __future__ import annotations

import contextlib
import sys
from pathlib import Path
from pprint import pformat
from typing import AsyncIterator, Self

import attrs
import click

from ai.backend.common import redis_helper
from ai.backend.common.config import redis_config_iv
from ai.backend.common.defs import REDIS_IMAGE_DB, REDIS_LIVE_DB, REDIS_STAT_DB, REDIS_STREAM_DB
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.exception import ConfigurationError
from ai.backend.common.logging import AbstractLogger, LocalLogger
from ai.backend.common.types import RedisConnectionInfo

from ..config import LocalConfig, SharedConfig
from ..config import load as load_config


class CLIContext:
    _local_config: LocalConfig | None
    _logger: AbstractLogger

    def __init__(self, config_path: Path, log_level: str) -> None:
        self.config_path = config_path
        self.log_level = log_level
        self._local_config = None

    @property
    def local_config(self) -> LocalConfig:
        # Lazy-load the configuration only when requested.
        try:
            if self._local_config is None:
                self._local_config = load_config(self.config_path, self.log_level)
        except ConfigurationError as e:
            print(
                "ConfigurationError: Could not read or validate the manager local config:",
                file=sys.stderr,
            )
            print(pformat(e.invalid_data), file=sys.stderr)
            raise click.Abort()
        return self._local_config

    def __enter__(self) -> Self:
        # The "start-server" command is injected by ai.backend.cli from the entrypoint
        # and it has its own multi-process-aware logging initialization.
        # If we duplicate the local logging with it, the process termination may hang.
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            self._logger = LocalLogger({})
            self._logger.__enter__()
        return self

    def __exit__(self, *exc_info) -> None:
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            self._logger.__exit__()


@contextlib.asynccontextmanager
async def etcd_ctx(cli_ctx: CLIContext) -> AsyncIterator[AsyncEtcd]:
    local_config = cli_ctx.local_config
    creds = None
    if local_config["etcd"]["user"]:
        creds = {
            "user": local_config["etcd"]["user"],
            "password": local_config["etcd"]["password"],
        }
    scope_prefix_map = {
        ConfigScopes.GLOBAL: "",
        # TODO: provide a way to specify other scope prefixes
    }
    etcd = AsyncEtcd(
        local_config["etcd"]["addr"],
        local_config["etcd"]["namespace"],
        scope_prefix_map,
        credentials=creds,
    )
    try:
        yield etcd
    finally:
        await etcd.close()


@contextlib.asynccontextmanager
async def config_ctx(cli_ctx: CLIContext) -> AsyncIterator[SharedConfig]:
    local_config = cli_ctx.local_config
    # scope_prefix_map is created inside ConfigServer
    shared_config = SharedConfig(
        local_config["etcd"]["addr"],
        local_config["etcd"]["user"],
        local_config["etcd"]["password"],
        local_config["etcd"]["namespace"],
    )
    await shared_config.reload()
    raw_redis_config = await shared_config.etcd.get_prefix("config/redis")
    local_config["redis"] = redis_config_iv.check(raw_redis_config)
    try:
        yield shared_config
    finally:
        await shared_config.close()


@attrs.define(auto_attribs=True, frozen=True, slots=True)
class RedisConnectionSet:
    live: RedisConnectionInfo
    stat: RedisConnectionInfo
    image: RedisConnectionInfo
    stream: RedisConnectionInfo


@contextlib.asynccontextmanager
async def redis_ctx(cli_ctx: CLIContext) -> AsyncIterator[RedisConnectionSet]:
    local_config = cli_ctx.local_config
    shared_config = SharedConfig(
        local_config["etcd"]["addr"],
        local_config["etcd"]["user"],
        local_config["etcd"]["password"],
        local_config["etcd"]["namespace"],
    )
    await shared_config.reload()
    raw_redis_config = await shared_config.etcd.get_prefix("config/redis")
    local_config["redis"] = redis_config_iv.check(raw_redis_config)
    redis_live = redis_helper.get_redis_object(
        shared_config.data["redis"],
        name="mgr_cli.live",
        db=REDIS_LIVE_DB,
    )
    redis_stat = redis_helper.get_redis_object(
        shared_config.data["redis"],
        name="mgr_cli.stat",
        db=REDIS_STAT_DB,
    )
    redis_image = redis_helper.get_redis_object(
        shared_config.data["redis"],
        name="mgr_cli.image",
        db=REDIS_IMAGE_DB,
    )
    redis_stream = redis_helper.get_redis_object(
        shared_config.data["redis"],
        name="mgr_cli.stream",
        db=REDIS_STREAM_DB,
    )
    yield RedisConnectionSet(
        live=redis_live,
        stat=redis_stat,
        image=redis_image,
        stream=redis_stream,
    )
    await redis_stream.close()
    await redis_image.close()
    await redis_stat.close()
    await redis_live.close()



import enum


class ManagerStatus(str, enum.Enum):
    TERMINATED = "terminated"  # deprecated
    PREPARING = "preparing"  # deprecated
    RUNNING = "running"
    FROZEN = "frozen"


class SchedulerEvent(str, enum.Enum):
    SCHEDULE = "schedule"
    PREPARE = "prepare"
    SCALE_SERVICES = "scale_services"



"""remove unique constraint of endpoints.name

Revision ID: e812d42bc34f
Revises: 02535458c0b3
Create Date: 2023-09-06 19:16:50.916865

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "e812d42bc34f"
down_revision = "02535458c0b3"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint("uq_endpoints_name", "endpoints")


def downgrade():
    op.create_unique_constraint("uq_endpoints_name", "endpoints", ["name"])



"""add_startsat_field_into_kernels

Revision ID: da24ff520049
Revises: 529113b08c2c
Create Date: 2020-06-18 20:47:22.152831

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "da24ff520049"
down_revision = "529113b08c2c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("kernels", sa.Column("starts_at", sa.DateTime(timezone=True), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("kernels", "starts_at")
    # ### end Alembic commands ###



from __future__ import annotations

import json
import logging
import urllib.parse
from typing import Any, AsyncIterator, Mapping, Optional, cast

import aiohttp
import aiotools
import yarl

from ai.backend.common.docker import arch_name_aliases
from ai.backend.common.docker import login as registry_login
from ai.backend.common.logging import BraceStyleAdapter

from .base import (
    BaseContainerRegistry,
    concurrency_sema,
    progress_reporter,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class HarborRegistry_v1(BaseContainerRegistry):
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        api_url = self.registry_url / "api"
        registry_projects = self.registry_info["project"]
        rqst_args = {}
        if self.credentials:
            rqst_args["auth"] = aiohttp.BasicAuth(
                self.credentials["username"],
                self.credentials["password"],
            )
        project_list_url: Optional[yarl.URL]
        project_list_url = (api_url / "projects").with_query(
            {"page_size": "30"},
        )
        project_ids = []
        while project_list_url is not None:
            async with sess.get(project_list_url, allow_redirects=False, **rqst_args) as resp:
                projects = await resp.json()
                for item in projects:
                    if item["name"] in registry_projects:
                        project_ids.append(item["project_id"])
                project_list_url = None
                next_page_link = resp.links.get("next")
                if next_page_link:
                    next_page_url = cast(yarl.URL, next_page_link["url"])
                    project_list_url = self.registry_url.with_path(next_page_url.path).with_query(
                        next_page_url.query
                    )
        if not project_ids:
            log.warning("There is no given project.")
            return
        repo_list_url: Optional[yarl.URL]
        for project_id in project_ids:
            repo_list_url = (api_url / "repositories").with_query(
                {"project_id": project_id, "page_size": "30"},
            )
            while repo_list_url is not None:
                async with sess.get(repo_list_url, allow_redirects=False, **rqst_args) as resp:
                    items = await resp.json()
                    repos = [item["name"] for item in items]
                    for item in repos:
                        yield item
                    repo_list_url = None
                    next_page_link = resp.links.get("next")
                    if next_page_link:
                        next_page_url = cast(yarl.URL, next_page_link["url"])
                        repo_list_url = self.registry_url.with_path(next_page_url.path).with_query(
                            next_page_url.query
                        )

    async def _scan_tag(
        self,
        sess: aiohttp.ClientSession,
        rqst_args: dict[str, Any],
        image: str,
        tag: str,
    ) -> None:
        async with concurrency_sema.get():
            async with sess.get(
                self.registry_url / f"v2/{image}/manifests/{tag}", **rqst_args
            ) as resp:
                if resp.status == 404:
                    # ignore missing tags
                    # (may occur after deleting an image from the docker hub)
                    return
                resp.raise_for_status()
                data = await resp.json()

                config_digest = data["config"]["digest"]
                size_bytes = sum(layer["size"] for layer in data["layers"]) + data["config"]["size"]
                async with sess.get(
                    self.registry_url / f"v2/{image}/blobs/{config_digest}", **rqst_args
                ) as resp:
                    resp.raise_for_status()
                    data = json.loads(await resp.read())
                    architecture = arch_name_aliases.get(data["architecture"], data["architecture"])
                    labels = {}
                    if "container_config" in data:
                        raw_labels = data["container_config"].get("Labels")
                        if raw_labels:
                            labels.update(raw_labels)
                        else:
                            log.warn("label not found on image {}:{}/{}", image, tag, architecture)
                    else:
                        raw_labels = data["config"].get("Labels")
                        if raw_labels:
                            labels.update(raw_labels)
                        else:
                            log.warn("label not found on image {}:{}/{}", image, tag, architecture)
                    manifest = {
                        architecture: {
                            "size": size_bytes,
                            "labels": labels,
                            "digest": config_digest,
                        },
                    }
        await self._read_manifest(image, tag, manifest)


class HarborRegistry_v2(BaseContainerRegistry):
    async def fetch_repositories(
        self,
        sess: aiohttp.ClientSession,
    ) -> AsyncIterator[str]:
        api_url = self.registry_url / "api" / "v2.0"
        registry_projects = self.registry_info["project"]
        rqst_args = {}
        if self.credentials:
            rqst_args["auth"] = aiohttp.BasicAuth(
                self.credentials["username"],
                self.credentials["password"],
            )
        repo_list_url: Optional[yarl.URL]
        for project_name in registry_projects:
            repo_list_url = (api_url / "projects" / project_name / "repositories").with_query(
                {"page_size": "30"},
            )
            while repo_list_url is not None:
                async with sess.get(repo_list_url, allow_redirects=False, **rqst_args) as resp:
                    items = await resp.json()
                    if isinstance(items, dict) and (errors := items.get("errors", [])):
                        raise RuntimeError(
                            f"failed to fetch repositories in project {project_name}",
                            errors[0]["code"],
                            errors[0]["message"],
                        )
                    repos = [item["name"] for item in items]
                    for item in repos:
                        yield item
                    repo_list_url = None
                    next_page_link = resp.links.get("next")
                    if next_page_link:
                        next_page_url = cast(yarl.URL, next_page_link["url"])
                        repo_list_url = self.registry_url.with_path(next_page_url.path).with_query(
                            next_page_url.query
                        )

    async def _scan_image(
        self,
        sess: aiohttp.ClientSession,
        image: str,
    ) -> None:
        api_url = self.registry_url / "api" / "v2.0"
        rqst_args = await registry_login(
            sess,
            self.registry_url,
            self.credentials,
            f"repository:{image}:pull",
        )
        if self.credentials:
            rqst_args["auth"] = aiohttp.BasicAuth(
                self.credentials["username"],
                self.credentials["password"],
            )
        project, _, repository = image.partition("/")
        project, repository = [urllib.parse.urlencode({"": x})[1:] for x in [project, repository]]
        async with aiotools.TaskGroup() as tg:
            artifact_url: Optional[yarl.URL] = (
                api_url / "projects" / project / "repositories" / repository / "artifacts"
            ).with_query(
                {"page_size": "30"},
            )
            while artifact_url is not None:
                async with sess.get(artifact_url, allow_redirects=False, **rqst_args) as resp:
                    resp.raise_for_status()
                    body = await resp.json()
                    for image_info in body:
                        skip_reason: Optional[str] = None
                        tag = image_info["digest"]
                        try:
                            if not image_info["tags"] or len(image_info["tags"]) == 0:
                                skip_reason = "no tag"
                                return
                            tag = image_info["tags"][0]["name"]
                            match image_info["manifest_media_type"]:
                                case self.MEDIA_TYPE_OCI_INDEX:
                                    await self._process_oci_index(
                                        tg, sess, rqst_args, image, image_info
                                    )
                                case self.MEDIA_TYPE_DOCKER_MANIFEST_LIST:
                                    await self._process_docker_v2_multiplatform_image(
                                        tg, sess, rqst_args, image, image_info
                                    )
                                case self.MEDIA_TYPE_DOCKER_MANIFEST:
                                    await self._process_docker_v2_image(
                                        tg, sess, rqst_args, image, image_info
                                    )
                                case _ as media_type:
                                    raise RuntimeError(
                                        f"Unsupported artifact media-type: {media_type}"
                                    )
                        finally:
                            if skip_reason:
                                log.warning("Skipped image - {}:{} ({})", image, tag, skip_reason)
                    artifact_url = None
                    next_page_link = resp.links.get("next")
                    if next_page_link:
                        next_page_url = cast(yarl.URL, next_page_link["url"])
                        artifact_url = self.registry_url.with_path(next_page_url.path).with_query(
                            next_page_url.query
                        )

    async def _process_oci_index(
        self,
        tg: aiotools.TaskGroup,
        sess: aiohttp.ClientSession,
        _rqst_args: Mapping[str, Any],
        image: str,
        image_info: Mapping[str, Any],
    ) -> None:
        rqst_args = dict(_rqst_args)
        if not rqst_args.get("headers"):
            rqst_args["headers"] = {}
        rqst_args["headers"].update({"Accept": "application/vnd.oci.image.manifest.v1+json"})
        digests: list[tuple[str, str]] = []
        tag_name = image_info["tags"][0]["name"]
        for reference in image_info["references"]:
            if (
                reference["platform"]["os"] == "unknown"
                or reference["platform"]["architecture"] == "unknown"
            ):
                continue
            digests.append((reference["child_digest"], reference["platform"]["architecture"]))
        if (reporter := progress_reporter.get()) is not None:
            reporter.total_progress += len(digests)
        async with aiotools.TaskGroup() as tg:
            for digest, architecture in digests:
                tg.create_task(
                    self._harbor_scan_tag_per_arch(
                        sess,
                        rqst_args,
                        image,
                        digest=digest,
                        tag=tag_name,
                        architecture=architecture,
                    )
                )

    async def _process_docker_v2_multiplatform_image(
        self,
        tg: aiotools.TaskGroup,
        sess: aiohttp.ClientSession,
        _rqst_args: Mapping[str, Any],
        image: str,
        image_info: Mapping[str, Any],
    ) -> None:
        rqst_args = dict(_rqst_args)
        if not rqst_args.get("headers"):
            rqst_args["headers"] = {}
        rqst_args["headers"].update({
            "Accept": "application/vnd.docker.distribution.manifest.v2+json"
        })
        digests: list[tuple[str, str]] = []
        tag_name = image_info["tags"][0]["name"]
        for reference in image_info["references"]:
            if (
                reference["platform"]["os"] == "unknown"
                or reference["platform"]["architecture"] == "unknown"
            ):
                continue
            digests.append((reference["child_digest"], reference["platform"]["architecture"]))
        if (reporter := progress_reporter.get()) is not None:
            reporter.total_progress += len(digests)
        async with aiotools.TaskGroup() as tg:
            for digest, architecture in digests:
                tg.create_task(
                    self._harbor_scan_tag_per_arch(
                        sess,
                        rqst_args,
                        image,
                        digest=digest,
                        tag=tag_name,
                        architecture=architecture,
                    )
                )

    async def _process_docker_v2_image(
        self,
        tg: aiotools.TaskGroup,
        sess: aiohttp.ClientSession,
        _rqst_args: Mapping[str, Any],
        image: str,
        image_info: Mapping[str, Any],
    ) -> None:
        rqst_args = dict(_rqst_args)
        if not rqst_args.get("headers"):
            rqst_args["headers"] = {}
        rqst_args["headers"].update({
            "Accept": "application/vnd.docker.distribution.manifest.v2+json"
        })
        if (reporter := progress_reporter.get()) is not None:
            reporter.total_progress += 1
        tag_name = image_info["tags"][0]["name"]
        async with aiotools.TaskGroup() as tg:
            tg.create_task(
                self._harbor_scan_tag_single_arch(
                    sess,
                    rqst_args,
                    image,
                    tag=tag_name,
                )
            )

    async def _harbor_scan_tag_per_arch(
        self,
        sess: aiohttp.ClientSession,
        rqst_args: dict[str, Any],
        image: str,
        *,
        digest: str,
        tag: str,
        architecture: str,
    ) -> None:
        """
        Scan 'image:tag' when there are explicitly known values of digest and architecture.
        """
        manifests = {}
        async with concurrency_sema.get():
            async with sess.get(
                self.registry_url / f"v2/{image}/manifests/{digest}", **rqst_args
            ) as resp:
                if resp.status == 404:
                    # ignore missing tags
                    # (may occur after deleting an image from the docker hub)
                    return
                resp.raise_for_status()
                top_manifest = await resp.json()
            architecture = arch_name_aliases.get(architecture, architecture)
            config_digest = top_manifest["config"]["digest"]
            size_bytes = (
                sum(layer["size"] for layer in top_manifest["layers"])
                + top_manifest["config"]["size"]
            )
            async with sess.get(
                self.registry_url / f"v2/{image}/blobs/{config_digest}", **rqst_args
            ) as resp:
                resp.raise_for_status()
                data = json.loads(await resp.read())
            labels = {}
            if "container_config" in data:
                raw_labels = data["container_config"].get("Labels")
                if raw_labels:
                    labels.update(raw_labels)
                else:
                    log.warning(
                        "label not found on image {}:{}/{}",
                        image,
                        tag,
                        architecture,
                    )
            else:
                raw_labels = data["config"].get("Labels")
                if raw_labels:
                    labels.update(raw_labels)
                else:
                    log.warning(
                        "label not found on image {}:{}/{}",
                        image,
                        tag,
                        architecture,
                    )
            manifests[architecture] = {
                "size": size_bytes,
                "labels": labels,
                "digest": config_digest,
            }
            await self._read_manifest(image, tag, manifests)

    async def _harbor_scan_tag_single_arch(
        self,
        sess: aiohttp.ClientSession,
        rqst_args: dict[str, Any],
        image: str,
        tag: str,
    ) -> None:
        """
        Scan 'image:tag' which has been pusehd as a single architecture tag.
        In this case, Harbor does not provide explicit methods to determine the architecture.
        We infer the architecture from the tag naming patterns ("-arm64" for instance).
        """
        manifests = {}
        async with concurrency_sema.get():
            rqst_args["headers"]["Accept"] = self.MEDIA_TYPE_DOCKER_MANIFEST
            # Harbor does not provide architecture information for a single-arch tag reference.
            # We heuristically detect the architecture using the tag name pattern.
            if tag.endswith("-arm64") or tag.endswith("-aarch64"):
                architecture = "aarch64"
            else:
                architecture = "x86_64"
            async with sess.get(
                self.registry_url / f"v2/{image}/manifests/{tag}", **rqst_args
            ) as resp:
                data = await resp.json()
                config_digest = data["config"]["digest"]
                size_bytes = sum(layer["size"] for layer in data["layers"]) + data["config"]["size"]
            async with sess.get(
                self.registry_url / f"v2/{image}/blobs/{config_digest}", **rqst_args
            ) as resp:
                resp.raise_for_status()
                data = json.loads(await resp.read())
            labels = {}
            if "container_config" in data:
                raw_labels = data["container_config"].get("Labels")
                if raw_labels:
                    labels.update(raw_labels)
                else:
                    log.warning(
                        "label not found on image {}:{}/{}",
                        image,
                        tag,
                        architecture,
                    )
            else:
                raw_labels = data["config"].get("Labels")
                if raw_labels:
                    labels.update(raw_labels)
                else:
                    log.warning(
                        "label not found on image {}:{}/{}",
                        image,
                        tag,
                        architecture,
                    )
            manifests[architecture] = {
                "size": size_bytes,
                "labels": labels,
                "digest": config_digest,
            }
            await self._read_manifest(image, tag, manifests)



"""add use_host_network column

Revision ID: c53397a490be
Revises: 5bce905c21e5
Create Date: 2022-10-25 17:02:31.709513

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql.expression import false

# revision identifiers, used by Alembic.
revision = "c53397a490be"
down_revision = "5bce905c21e5"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "kernels",
        sa.Column(
            "use_host_network", sa.BOOLEAN(), nullable=False, server_default=false(), default=False
        ),
    )
    op.add_column(
        "scaling_groups",
        sa.Column(
            "use_host_network", sa.BOOLEAN(), nullable=False, server_default=false(), default=False
        ),
    )


def downgrade():
    op.drop_column("kernels", "use_host_network")
    op.drop_column("scaling_groups", "use_host_network")
